#include <sys/types.h>
#include <sys/resource.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <dirent.h>
#include <signal.h>
#include <popt.h>

#include "libprofile_ctrl.h"
#include "prof_config.h"

#define MAX_EVENTNUM 4
#define MAX_FNAME 256
#define NOTE_SIZE 16384
#define BUF_SIZE 65536

#define verbprintf(args...) 
//			printf(args) 

#define CA_BASE_DIR "/var/lib/ca_oprofile"
#define CA_LOCK_FILE "/var/lib/ca_oprofile/lock"


sig_atomic_t signal_alarm;
sig_atomic_t signal_hup;
sig_atomic_t signal_term;
sig_atomic_t signal_usr1;
sig_atomic_t signal_usr2;


uint op_nr_counters;
int verbose;
int separate_lib;
int separate_kernel;
int separate_thread;
int separate_cpu;
int no_vmlinux;
char * vmlinux;
char * kernel_range;
static char * binary_name_filter;
static char * events;
static int showvers;
static struct event_properties ca_events;
		

static struct poptOption options[] = {
	{ "kernel-range", 'r', POPT_ARG_STRING, &kernel_range, 0, "Kernel VMA range", "start-end", },
	{ "vmlinux", 'k', POPT_ARG_STRING, &vmlinux, 0, "vmlinux kernel image", "file", },
	{ "no-vmlinux", 0, POPT_ARG_NONE, &no_vmlinux, 0, "vmlinux kernel image file not available", NULL, },
	{ "image", 0, POPT_ARG_STRING, &binary_name_filter, 0, "image name filter", "profile these comma separated image" },
	{ "separate-lib", 0, POPT_ARG_INT, &separate_lib, 0, "separate library samples for each distinct application", "[0|1]", },
	{ "separate-kernel", 0, POPT_ARG_INT, &separate_kernel, 0, "separate kernel samples for each distinct application", "[0|1]", },
	{ "separate-thread", 0, POPT_ARG_INT, &separate_thread, 0, "thread-profiling mode", "[0|1]" },
	{ "separate-cpu", 0, POPT_ARG_INT, &separate_cpu, 0, "separate samples for each CPU", "[0|1]" },
	{ "events", 'e', POPT_ARG_STRING, &events, 0, "events list", "[events]" },
	{ "version", 'v', POPT_ARG_NONE, &showvers, 0, "show version", NULL, },
	{ "verbose", 'V', POPT_ARG_NONE, &verbose, 0, "be verbose in log file", NULL, },
	POPT_AUTOHELP
	{ NULL, 0, 0, NULL, 0, NULL, NULL, },
};


static int cad_init_driver(void);
static int cad_stop_sampling(void);
 
char * op_xstrndup(char const * s, size_t len)
{
	return xmemdup(s, len, len + 1);
}


static void malformed_events(void)
{
	fprintf(stderr, "oprofiled: malformed events passed "
	        "on the command line\n");
	exit(EXIT_FAILURE);
}


static char * copy_token(char ** c, char delim)
{
	char * tmp = *c;
	char * tmp2 = *c;
	char * str;

	if (!**c)
		return NULL;

	while (*tmp2 && *tmp2 != delim)
		++tmp2;

	if (tmp2 == tmp)
		return NULL;

	str = op_xstrndup(tmp, tmp2 - tmp);
	*c = tmp2;
	if (**c)
		++*c;
	return str;
}


int hexalpha_to_int(int c)
{
  char hexalpha[] = "aAbBcCdDeEfF";
  int i;
  int answer = 0;

  for(i = 0; answer == 0 && hexalpha[i] != '\0'; i++)
  {
    if(hexalpha[i] == c)
    {
      answer = 10 + (i / 2);
    }
  }

  return answer;
}

unsigned int htol(const char s[])
{
  unsigned long answer = 0;
  int i = 0;
  int valid = 1;
  int hexit;

  if(s[i] == '0')
  {
    ++i;
    if(s[i] == 'x' || s[i] == 'X')
    {
      ++i;
    }
  }

  while(valid && s[i] != '\0')
  {
    answer = answer * 16;
    if(s[i] >= '0' && s[i] <= '9')
    {
      answer = answer + (s[i] - '0');
    }
    else
    {
      hexit = hexalpha_to_int(s[i]);
      if(hexit == 0)
      {
        valid = 0;
      }
      else
      {
        answer = answer + hexit;
      }
    }

    ++i;
 } 

  if(!valid)
  {
    answer = 0;
  }

  return answer;
}

static unsigned long copy_ulong(char ** c, char delim)
{
	unsigned long val = 0;
	char * str = copy_token(c, delim);
	if (!str)
		malformed_events();
	val = strtoul(str, NULL, 0);
	free(str);
	return val;
}


void cad_parse_kernel_range(char const * _range)
{
    char * range = xstrdup(_range);
    char * c;
    unsigned long kernel_begin = 0;
    unsigned long kernel_end = 0;

    if (!range || !strlen(range)) {
        fprintf(stderr, "cadaemon: no kernel range passed.\n");
		exit(EXIT_FAILURE);
    }

    verbprintf("Kernel range string: %s\n", range);

    c = range;

    char delim = ',';
	char * str = copy_token(&c, delim);

    kernel_begin = htol(str);
    kernel_end = htol(c);

    verbprintf("kernel range: %lx to %lx\n", 
        kernel_begin, kernel_end);
    
}


void opd_parse_events(char const * events)
{
	char * ev = xstrdup(events);
	char * c;
    char * name;
	size_t cur = 0;
    unsigned long kernel;
    unsigned long user;

	if (!ev || !strlen(ev)) {
		fprintf(stderr, "oprofiled: no events passed.\n");
		exit(EXIT_FAILURE);
	}

	verbprintf("Events: %s\n", ev);

	c = ev;

	while (*c && cur < op_nr_counters) {
		struct event * e = (struct event *)&ca_events.event_value[cur];

        if (e != NULL) 
        {
            name = copy_token(&c, ':');

            e->event_index = copy_ulong(&c, ':');
            e->counter_mask = copy_ulong(&c, ':');
            ca_events.event_count[cur] = copy_ulong(&c, ':');
            e->unit_mask = copy_ulong(&c, ':');
            if (kernel = copy_ulong(&c, ':'))
	            e->bit_mask |= ESCR_OS;
            if (user = copy_ulong(&c, ':'))
	            e->bit_mask |= ESCR_USER;
       } 

		++cur;
	}

	if (*c) {
		fprintf(stderr, "oprofiled: too many events passed.\n");
		exit(EXIT_FAILURE);
	}

	free(ev);

	//cpu_speed = op_cpu_frequency();
}


/**
 * opd_alarm - sync files and report stats
 */
static void opd_alarm(int val __attribute__((unused)))
{
	signal_alarm = 1;
}


/* re-open logfile for logrotate */
static void opd_sighup(int val __attribute__((unused)))
{
	signal_hup = 1;
}


static void opd_sigterm(int val __attribute__((unused)))
{
    cad_stop_sampling();
	signal_term = 1;
}
 

static void opd_sigusr1(int val __attribute__((unused)))
{
    cad_start_sampling();
    signal_usr1 = 1;
}

 
static void opd_sigusr2(int val __attribute__((unused)))
{
    cad_stop_sampling();
	signal_usr2 = 1;
}


static void opd_setup_signals(void)
{
    struct sigaction act;
 
	act.sa_handler = opd_alarm;
	act.sa_flags = 0;
	sigemptyset(&act.sa_mask);

	if (sigaction(SIGALRM, &act, NULL)) {
		perror("oprofiled: install of SIGALRM handler failed: ");
		exit(EXIT_FAILURE);
	}

	act.sa_handler = opd_sighup;
	act.sa_flags = 0;
	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGALRM);

	if (sigaction(SIGHUP, &act, NULL)) {
		perror("oprofiled: install of SIGHUP handler failed: ");
		exit(EXIT_FAILURE);
	}

	act.sa_handler = opd_sigterm;
	act.sa_flags = 0;
	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGTERM);

	if (sigaction(SIGTERM, &act, NULL)) {
		perror("oprofiled: install of SIGTERM handler failed: ");
		exit(EXIT_FAILURE);
	}

	act.sa_handler = opd_sigusr1;
	act.sa_flags = 0;
	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGTERM);

	if (sigaction(SIGUSR1, &act, NULL)) {
		perror("oprofiled: install of SIGUSR1 handler failed: ");
		exit(EXIT_FAILURE);
	}

	act.sa_handler = opd_sigusr2;
	act.sa_flags = 0;
	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGTERM);

	if (sigaction(SIGUSR2, &act, NULL)) {
		perror("oprofiled: install of SIGUSR1 handler failed: ");
		exit(EXIT_FAILURE);
	}

	/* clean up every 10 minutes */
	//alarm(60 * 1);
}


static void opd_options(int argc, char const * argv[])
{
	poptContext optcon;
    int c = 0;


	optcon = poptGetContext(NULL, argc, argv, options, 0);

    while (c > -1)
    {
	    c = poptGetNextOpt(optcon);
    }

	//if (showvers)
    //		show_version(argv[0]);

    //todo: use the #defined  
	op_nr_counters = MAX_EVENTNUM;

    /*
	if (!no_vmlinux) {
		if (!vmlinux || !strcmp("", vmlinux)) {
			fprintf(stderr, "oprofiled: no vmlinux specified.\n");
			poptPrintHelp(optcon, stderr, 0);
			exit(EXIT_FAILURE);
		}
        */

		/* canonicalise vmlinux filename. fix #637805 */
		//vmlinux = op_relative_to_absolute_path(vmlinux, NULL);
        /*

		if (!kernel_range || !strcmp("", kernel_range)) {
			fprintf(stderr, "oprofiled: no kernel VMA range specified.\n");
			poptPrintHelp(optcon, stderr, 0);
			exit(EXIT_FAILURE);
		}
    }	
    */

    if (!no_vmlinux) {
        cad_parse_kernel_range(kernel_range);
    }

	opd_parse_events(events);


	poptFreeContext(optcon);
}


/**
 * opd_fork - fork and return as child
 *
 * fork() and exit the parent with _exit().
 * Failure is fatal.
 */
static void opd_fork(void)
{
    printf("in opd_fork");
	switch (fork()) {
		case -1:
			perror("oprofiled: fork() failed: ");
			exit(EXIT_FAILURE);
			break;
		case 0:
            printf("child\n");
			break;
		default:
			/* parent */
			_exit(EXIT_SUCCESS);
			break;
	}
}

 
static void opd_go_daemon(void)
{
    printf("in opd_go_daemon\n");
	opd_fork();

	if (chdir(CA_BASE_DIR)) {
		fprintf(stderr,"oprofiled: opd_go_daemon: couldn't chdir to "
			CA_BASE_DIR ": %s", strerror(errno));
		exit(EXIT_FAILURE);
	}

	if (setsid() < 0) {
		perror("oprofiled: opd_go_daemon: couldn't setsid: ");
		exit(EXIT_FAILURE);
	}

	opd_fork();
}


static int cad_set_events(void)
{
    return profile_set_event(&ca_events);
}

static int cad_init_driver(void)
{
    int val = 0;
    char output_name[MAX_FNAME];

	profile_init();
    
    memset(output_name, 0, MAX_FNAME);
    sprintf(output_name, "%s/%s", CA_BASE_DIR, "commandline");

    do { 
        if (-1 == (val = profile_set_note_buffer_size(NOTE_SIZE))) 
            break;

        if (-1 == (val = profile_set_sample_buffer_size(BUF_SIZE)))
            break;

        verbprintf("output_name %s\n", output_name);
        if (-1 == (val = profile_set_output(output_name, strlen(output_name))))
            break;
    } while (0);

    return val;
}

static int cad_start_sampling(void)
{
    verbprintf("Start sampling\n");
    return profile_start();
}

static int cad_stop_sampling(void)
{
    verbprintf("Stop sampling\n");
    return profile_stop();
}

int main(int argc, char const * argv[])
{
    do {
	    opd_setup_signals();
	
		opd_options(argc, argv);
	
	    if (0 != cad_init_driver())
            break;
	
	    if (0 != cad_set_events())
            break;
	

	    opd_go_daemon();

		if (op_write_lock_file(CA_LOCK_FILE)) {
			fprintf(stderr, "oprofiled: could not create lock file "
				CA_LOCK_FILE "\n");
			exit(EXIT_FAILURE);
		}
	
        // Not the best way to implement it.  Use 
        // sleep() for now.
        while(1){
            sleep(1);
        }
    } while(0);

    if (0 != unlink(CA_LOCK_FILE)) {
        verbprintf("failed on unlink\n");
    }
    verbprintf("existing\n");
}
